# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


class SmtpConnectionProperties:
    """
    Represent the properties of connection to a Simple Mail Transport
    Protocol (SMTP) server.
    """
    def __eq__(self, other):
        return self.hostname.lower() == other.hostname.lower() \
            and self.port_number == other.port_number \
            and self.username == other.username \
            and self.password == other.password

    def __init__(
            self,
            hostname,
            username,
            password,
            port_number=587):
        """
        Build the connection properties to access a SMTP server

        :param hostname: Internet address or fully qualified domain name --
            human-readable nickname that corresponds to the address -- of the
            SMTP server is running on.

        :param port_number:  Internet port number on which the remote SMTP
            server is listening at.

            SMTP communication between mail servers uses TCP port 25.  Mail
            clients on the other hand, often submit the outgoing emails to a
            mail server on port 587.  Despite being deprecated, mail providers
            sometimes still permit the use of nonstandard port 465 for this
            purpose.

        :param username: username to authenticate with against the SMTP server.

        :param password: password associate to the username to authenticate
            with against the SMTP server.
        """
        self.hostname = hostname
        self.port_number = port_number
        self.username = username
        self.password = password



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re


# Regular expression that matches a generic international phone number.
# US land line centric international access code 011 is replaced with
# the standard international access code identifier of '+', making it
# mandatory. The minimum for the national number is at least one digit.
#
# Note that if you enter numbers in this format into your mobile phone
# address book, you may successfully call any number in your address
# book no matter where you travel. For land lines, replace the plus with
# the international access code for the country you are dialing from.
#
# Note that this DOES NOT take into account national number plan rules
# - specifically, it allows zeros and ones in locations that national
# number plans may not allow and also allows number lengths greater than
# the national number plan for some countries (e.g., the US).
REGEX_E164_PHONE_NUMBER = re.compile(r"^\+((?:9[679]|8[035789]|6[789]|5[90]|42|3[578]|2[1-689])|9[0-58]|8[1246]|6[0-6]|5[1-8]|4[013-9]|3[0-469]|2[70]|7|1)(?:\W*\d){0,13}\d$")



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import errno
import gzip
import os
import zipfile

try:
    import bz2
except ImportError as error:
    bz2 = None

from logging.handlers import RotatingFileHandler
#try:
#    from cloghandler import ConcurrentRotatingFileHandler as RotatingFileHandler
#except ImportError:
#    warn(__file__)
#    warn("ConcurrentLogHandler package not installed; using built-in log handler")
#    from logging.handlers import RotatingFileHandler


class RotatingFileHandlerEx(RotatingFileHandler):
    """
    Provides an additional log handler for Python's standard logging
    package (PEP 282).  This handler will write log events to log file
    which is rotated when the log file reaches a certain size.  The log
    file will be compressed.  Multiple processes can safely write to the
    same log file concurrently.
    """
    def __init__(self, path, maxBytes, backupCount,
            compression='gzip'):
        """
        Build a new instance of the ``RotatingFileHandlerEx`` class.


        @param path: absolute path name where the rotating log file will be
            written into.

        @param maxBytes: maximum number of bytes to be written before rolling
            over the current log file, compressing it, and creating a new log
            file.

        @param backupCount: if non-zero, the system will save old log files by
            appending the extensions ``.1``, ``.2`` etc., to the filename.  For
            example, with a ``backupCount`` of ``5`` and a base file name of
            ``app.log``, the logger generates ``app.log``, ``app.log.1``,
            ``app.log.2``, up to ``app.log.5``.  The file being written to is
            always ``app.log``.  When this file is filled, it is closed and
            renamed to ``app.log.1``, and if files ``app.log.1``, ``app.log.2``,
            etc. exist, then they are renamed to ``app.log.2``, ``app.log.3``,
            etc. respectively.

        @param compression: indicate which compression has to be used:

            * ``none``: NoneCompressor

            * ``zip``: ZipCompressor

            * ``gzip``: GzipCompressor

            * ``bzip2``: Bzip2Compressor
        """
        print('[INFO] Writing logs into %s (%d MB, %d files)' % (path, maxBytes / 1024 / 1024, backupCount))

        filename = os.path.abspath(path)
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise exception

        super(self.__class__, self).__init__(
            filename, mode='a', maxBytes=maxBytes, backupCount=backupCount)
        self.baseFilename = filename
        self.compressor = get_compressor(compression)

    def doRollover(self):
        self.stream.close()
        if self.backupCount > 0:
            file_pattern = self.compressor.file_pattern
            for i in range(self.backupCount - 1, 0, -1):
                sfn = file_pattern % (self.baseFilename, i)
                dfn = file_pattern % (self.baseFilename, i + 1)
                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            self.compressor.compress(self.baseFilename)
        self.stream = open(self.baseFilename, "w")

    def reopen(self):
        self.close()
        self.stream = open(self.baseFilename, 'a')


class NoneCompressor(object):
    file_pattern = '%s.%d'

    def __init__(self):
        pass
        # warn("[WARNING] Using no compression for logging.")

    def compress(self, filename):
        dfn = filename + '.1'
        if os.path.exists(dfn):
            os.remove(dfn)
        os.rename(filename, dfn)


class ZipCompressor(object):
    file_pattern = '%s.%d.zip'

    def compress(self, filename):
        dfn = filename + '.1.zip'
        if os.path.exists(dfn):
            os.remove(dfn)
        zf = zipfile.ZipFile(dfn, 'w', zipfile.ZIP_DEFLATED)
        zf.write(filename, os.path.basename(filename))
        zf.close()
        os.remove(filename)


class GzipCompressor(object):
    file_pattern = '%s.%d.gz'

    def compress(self, filename):
        dfn = filename + '.1.gz'
        if os.path.exists(dfn):
            os.remove(dfn)
        zf = gzip.GzipFile(dfn, 'wb')
        zf.write(open(filename, 'rb').read())
        zf.close()
        os.remove(filename)


if bz2:
    class Bzip2Compressor(object):
        file_pattern = '%s.%d.bz2'

        def compress(self, filename):
            dfn = filename + '.1.bz2'
            if os.path.exists(dfn):
                os.remove(dfn)
            zf = bz2.BZ2File(dfn, 'w')
            zf.write(open(filename, 'rb').read())
            zf.close()
            os.remove(filename)
else:
    Bzip2Compressor = NoneCompressor


def get_compressor(compression):
    # { compression mode: compressor class, ...}
    compressors = {
        'none': NoneCompressor,
        'zip': ZipCompressor,
        'gzip': GzipCompressor,
        'bzip2': Bzip2Compressor
    }

    klass = compressors.get(compression, NoneCompressor)
    return klass()



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import importlib
import sys
from os import PathLike
from pathlib import Path


def get_project_root_path(
        module_file_path_name: PathLike,
        module_name: str,
        source_depth: int = 0
) -> PathLike:
    """
    Return the root path of the Python project.


    :note: The function detects whether the module has been installed as a
        3rd party library (i.e., within the directory ``site-packages``).
        In this case, any source directory would have been removed from
        the path to the module file.  Therefore, the argument
        ``source_depth`` is automatically set to ``0``.


    :param module_file_path_name: The path to the module from which this
        function is called.  This SHOULD be the Python built-in variable
        ``__file__``, which evaluates to the pathname of the current
        module.

    :param module_name: The name of the module from which this function is
        called.  This SHOULD be the Python built-in variable ``__name__``,
        which evaluates to the name of the current module.

    :param source_depth: The depth of the Python project's source directory
        from the Python project's root directory.

        Python developers usually create a ``src`` directory at the root of
        their Python project where they organize their project's modules,
        rather than putting their sources file directly in the root
        directory.  For example::

        ```text
        .
        ├── CHANGELOG.md
        ├── LICENSE.md
        ├── README.md
        ├── data
        │   ├── ...
        │   .
        │   └── ...
        ├── poetry.lock
        ├── pyproject.toml
        ├── run.py
        └── src
            └── ...
                ├── __init__.py
                ├── ...
                .
                └── ...
                .   ├── __init__.py
                .   .
                .   └── ...
                └── ...
        ```

        Their ``pyproject.toml`` file contains the following instruction:

        ```toml
        packages = [{ include = "majormode", from = "src" }]
        ```

        In this case, the depth of the Python project's source directory
        is ``1``.

        This argument is automatically sets to ``0`` if the module has
        been installed as 3rd party library, in which case Python project's
        source directory has been removed from the path to this module file.


    :return: The root path of the Python project.


    :raise ValueError: If the module name doesn't refer to a Python
        package but to a not packaged script (i.e., ``__main__``).
    """
    if module_name == '__main__':
        raise ValueError("A Python packaged module is required")

    if source_depth < 0:
        raise ValueError("The source depth must be a positive integer")

    # Check whether the Python module has been installed as a 3rd party
    # library, i.e., inside the directory `site-packages`.  This directory
    # is where user installed packages are dropped.  In this case, the
    # argument `source_depth` is not relevant.
    if 'site-packages' in str(module_file_path_name):
        source_depth = 0

    # Determine the number of subpackages that compose the module name (also
    # known as dotted module names).
    #
    # For instance, the file `{{PROJECT_PATH}}/src/majormode/perseus/utils/logging.py`
    # corresponds to the module `majormode.perseus.utils.logging`.  This
    # module is composed of 4 subpackages:
    #
    # ```python
    # ['majormode', 'perseus', 'utils', 'logging']
    # ```
    #
    # If the module belongs to the current Python project, the path to the
    # `logging.py` file  is for example:
    #
    # ```path
    # /User/foobar/Devel/Perseus/perseus-core-python-library/src/majormode/perseus/utils/
    # ```
    #
    # The project path is determined from the **logical ancestors** of the
    # `logging.py` path (i.e., from the right to the left) with the
    # following Python expression:
    #
    # ```python
    # Path(module_file_path_name).parents[
    #     # Exclude the `logging.py` file from the subpackage count, as it is
    #     # not part of the directory path.
    #     module_package_depth - 1
    #
    #     # Excluding any first directories where the package is stored inside.
    #     + source_depth
    # ]
    # ```
    #
    # WARNING: SPECIAL CASE WITH THE `__init__.py` FILE!  For example, the
    # file ``{{PROJECT_PATH}}/src/majormode/perseus/utils/__init__.py``
    # corresponds to the module `majormode.perseus.utils`.  This module is
    # composed of 3 subpackages:
    #
    # ```python
    # ['majormode', 'perseus', 'utils']
    # ```
    #
    # However, the path to the `__init__.py` file remains the same that in
    # the previous example:
    #
    # ```path
    # /User/foobar/Devel/Perseus/perseus-core-python-library/src/majormode/perseus/utils/
    # ```
    #
    # The project path is determined with the following Python expression:
    #
    # ```python
    # Path(module_file_path_name).parents[
    #     # The `__init__.py` file was not counted as a subpackage.
    #     module_package_depth - 1 + 1
    #
    #     # Excluding any first directories where the package is stored inside.
    #     + source_depth
    # ]
    # ```
    module_package_depth = len(module_name.split('.'))

    if Path(module_file_path_name).name != '__init__.py':  # See comments above
        module_package_depth -= 1

    module_package_file_name = Path(module_file_path_name)
    project_path = module_package_file_name.parents[module_package_depth + source_depth]

    return project_path


def load_module(module_package_name):
    """
    Load the specified Python module specified by a string representation.


    @note: The function prevents from circular import.


    :param module_package_name: A string representation of the Python
        packaged module (module namespace) using "dotted module names",
        e.g., `foo.bar`.


    :return: An object of the specified module.
    """
    return sys.modules.get(module_package_name) or importlib.import_module(module_package_name)


def load_class(class_module_package_name):
    """
    Dynamically load a Python class specified by the string representation
    of this class in a Python packaged module (module namespace).


    @note: The function prevents from circular import.


    :param class_module_package_name: A string representation of the class
        to load using "dotted module names", e.g., `foo.bar.MyClass`.


    :return: The specified class.
    """
    module_package_name, class_name = class_module_package_name.rsplit('.', 1)

    return getattr(load_module(module_package_name), class_name)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import re

from majormode.perseus.constant.locale import ISO_3166_1_ALPHA_2_CODES
from majormode.perseus.constant.locale import ISO_639_1_CODES
from majormode.perseus.constant.locale import ISO_639_1_CODES_TO_ISO_639_3_CODES
from majormode.perseus.constant.locale import ISO_639_3_CODES


REGEX_PATTERN_LANGUAGE_CODE = r'[a-z]{2,3}'
REGEX_PATTERN_COUNTRY_CODE = r'[A-Z]{2}'
REGEX_PATTERN_LOCALE = r'(([a-z]{2,3})-([A-Z]{2})$)|([a-z]{2,3}$)'
REGEX_PATTERN_JAVA_LOCALE = r'(([a-z]{2})-([A-Z]{2})$)|([a-z]{2}$)'
REGEX_PATTERN_PERMISSIVE_LOCALE = r'(([a-z]{2,3})[-_]([A-Za-z]{2})$)|([a-z]{2,3}$)'

REGEX_LOCALE = re.compile(REGEX_PATTERN_LOCALE)
REGEX_JAVA_LOCALE = re.compile(REGEX_PATTERN_JAVA_LOCALE)
REGEX_PERMISSIVE_LOCALE = re.compile(REGEX_PATTERN_PERMISSIVE_LOCALE)


class Locale:
    """
    Represent a locale that corresponds to a tag respecting RFC 4646.

    Some computational tasks require information about the current user
    context to be able to process data—particularly when formatting output
    for presentation to the user or when interpreting input.  A locale
    object provides a repository for that information.  An operation that
    requires a locale object to perform its task is called locale-
    sensitive.

    A locale is not a language; it’s a set of conventions for handling
    written language text and various units (for example, date and time
    formats, currency used, and the decimal separator).

    Conceptually, a locale identifies a specific user community—a group of
    users who have similar cultural and linguistic expectations for human-
    computer interaction (and the kinds of data they process). A locale’s
    identifier is a label for a given set of settings.  For example, "en”
    (representing "English”) is an identifier for a linguistic (and to
    some extent cultural) locale that includes (among others) Australia,
    Great Britain, and the United States.  There are also specific
    regional locales for Australian English, British English, U.S.
    English, and so on.

    When data are displayed to a user it should be formatted according to
    the conventions of the user’s native country, region, or culture.
    Conversely, when users enter data, they may do so according to their
    own customs or preferences.  Locale objects are used to provide
    information required to localize the presentation or interpretation of
    data.  This information can include decimal separators, date formats,
    and units of measurement, as well as language and region information.

    Locales are arranged in a hierarchy. At the root is the system locale,
    which provides default values for all settings. Below the root
    hierarchy are language locales.  These encapsulate settings for
    language groups, such as English, German and Chinese (using
    identifiers "en”, "de”, and "zh”).  Normal locales specify a language
    in a particular region (for example "en-GB”, "de-AT”, and "zh-SG”).

    A locale is expressed by a ISO 639-3 alpha-3 code element, optionally
    followed by a dash character `-` and a ISO 3166-1 alpha-2 code.  For
    example: "eng" (which denotes a standard English), "eng-US" (which
    denotes an American English).
    """
    class MalformedLocaleException(Exception):
        """
        Indicate that a string doesn't comply with the valid expression of a
        locale.
        """

    class InvalidCountryCodeException(Exception):
        """
        Indicate that a string doesn't match a valid country code.
        """

    class InvalidLanguageCodeException(Exception):
        """
        Indicate that a string doesn't match a valid language code.
        """

    @classmethod
    def assert_country_code(cls, code: str, strict: bool = True) -> None:
        """
        Asset that a country code is valid


        :param code: A country code.

        :param strict: Indicate whether the comparison of the country code
            string is case-sensitive or not.


        :raise InvalidCountryCodeException: If the country code passed to this
            function doesn't match a valid language code.
        """
        if not cls.is_country_code(code, strict=strict):
            raise cls.InvalidCountryCodeException(f'Invalid country code "{code}"')

    @classmethod
    def assert_language_code(cls, code: str, strict: bool = True) -> None:
        """
        Assert the a language code is valid


        :param code: A language code.

        :param strict: Indicate whether the comparison of the language code
            string is case-sensitive or not.


        :raise InvalidLanguageCodeException: If the language code passed to
            this function doesn't match a valid language code.
        """
        if not cls.is_language_code(code, strict=strict):
            raise cls.InvalidLanguageCodeException(f'Invalid language code "{code}"')

    def __eq__(self, other: object) -> bool:
        """
        Indicate whether this locale is the same as the other locale.

        Two locale objects are the same when they correspond to the same
        language and the same country (unless no country is defined for both
        locales).


        :param other: The other locale to compare.


        :return: Indicate whether this locale is the identical to the other
            locale.
        """
        if not isinstance(other, Locale):
            return False

        return self.language_code == other.language_code and \
            self.__country_code == other.country_code

    def __hash__(self) -> int:
        if not hasattr(self, '__hash'):
            encoded_locale = self.language_code if self.__country_code is None \
                else f'{self.language_code}{self.__country_code}'

            self.__hash = sum([
                (ord(c) - ord('a') if ord(c) >= ord('a') else ord(c) - ord('A')) * 52**i
                for i, c in enumerate(encoded_locale)
            ])

        return self.__hash

    def __init__(
            self,
            language_code: str,
            country_code: str = None,
            strict: bool = True):
        """
        Build a locale providing a ISO 639-3 alpha-3 code (or alpha-2 code),
        and an optional ISO 3166-1 alpha-2 code


        :param language_code: A ISO 639-3 alpha-3 code (or alpha-2 code; which
            will be automatically converted to its equivalent ISO 639-3
            alpha-3 code).

        :param country_code: A ISO 3166-1 alpha-2 code.


        :raise InvalidCountryCodeException: If the argument `country_code`
            doesn't match a valid country_code code.

        :raise InvalidLanguageCodeException: If the argument `language_code`
            doesn't match a valid language code.
        """
        self.assert_language_code(language_code, strict=strict)
        if country_code:
            self.assert_country_code(country_code, strict=strict)

        self.__language_code = self.__to_iso_639_3(language_code)
        self.__country_code = country_code

    def __repr__(self) -> str:
        return self.to_string()

    def __str__(self) -> str:
        return self.to_string()

    @staticmethod
    def __to_iso_639_3(code: str) -> str:
        return code if len(code) == 3 else ISO_639_1_CODES_TO_ISO_639_3_CODES[code]

    @property
    def country_code(self) -> str:
        return self.__country_code

    @staticmethod
    def from_string(locale: str, strict: bool = True) -> Locale | None:
        """
        Return an object `Locale` corresponding to the string representation
        of a locale.


        :param locale: A string representation of a locale, i.e., a ISO 639-3
           alpha-3 code (or alpha-2 code), optionally followed by a dash
           character `-` and a ISO 3166-1 alpha-2 code.

        :param strict: Indicate whether the string representation of a locale
            has to be strictly compliant with RFC 4646, or whether a Java-style
            locale (character `_` instead of `-`) is accepted.


        :return: A locale or `None` if the argument `locale` is undefined.
        """
        if not locale:
            return None

        language_code, country_code = Locale.decompose_locale(locale, strict)
        return Locale(language_code, country_code, strict=strict)

    @classmethod
    def is_country_code(cls, code: str, strict: bool = True) -> bool:
        return cls.is_iso_3166_1_alpha_2(code, strict=strict)

    @staticmethod
    def is_iso_3166_1_alpha_2(code: str, strict: bool = True) -> bool:
        return (code if strict else code.upper()) in ISO_3166_1_ALPHA_2_CODES

    @staticmethod
    def is_iso_639_2_code(code: str, strict: bool = True) -> bool:
        return (code if strict else code.lower()) in ISO_639_1_CODES

    @staticmethod
    def is_iso_639_3_code(code: str, strict: bool = True):
        return (code if strict else code.lower()) in ISO_639_3_CODES

    @classmethod
    def is_language_code(cls, code: str, strict: bool = True):
        if 2 <= len(code) <= 3:
            return cls.is_iso_639_2_code(code, strict=strict) if len(code) == 2 \
                else cls.is_iso_639_3_code(code, strict=strict)
        return False

    def is_similar(self, other: Locale) -> bool:
        """
        Indicate whether the current locale object is similar to another
        passed to the comparison method.

        Two locale objects are said similar if they have at least the same
        language, but not necessarily the same country.


        :param other: a `Locale` object to compare with the current locale
            object.


        :return: `True` if the given locale is similar to the current
            locale; `False` otherwise.
        """
        return self.language_code == other.language_code

    @property
    def language_code(self) -> str:
        return self.__language_code
    
    def to_http_string(self) -> str:
        """
        Return the string representation of the locale compatible with the
        HTTP header `Accept-Language` as specified in `RFC 7231
        <https://tools.ietf.org/html/rfc7231#section-5.3.5>_`

        The Accept-Language request HTTP header advertises which languages the
        client is able to understand, and which locale variant is preferred.


        :return: a string representation of this locale compatible with HTTP
            request, i.e., a ISO 639-3 alpha-2, optionally followed by a dash
            character `-` and a ISO 3166-1 alpha-2 code.
        """
        return self.language_code[:2] if self.__country_code is None \
            else f'{self.language_code[:2]}-{self.__country_code}'

    def to_string(self) -> str:
        """
        Return a string representation of this object `Locale`.


        :return: a string representation of a locale, i.e., a ISO 639-3
             alpha-3 code (or alpha-2 code), optionally followed by a dash
            character `-` and a ISO 3166-1 alpha-2 code.
        """
        return Locale.compose_locale(self.language_code, self.__country_code)

    @staticmethod
    def decompose_locale(locale: str, strict: bool = True) -> (str, str):
        """
        Return the decomposition of the specified locale into a language
        and a country codes


        :param locale: A ISO 639-3 alpha-3 code (or alpha-2 code), optionally
            followed by a dash character `-` and a ISO 3166-1 alpha-2 code.
            If `None` passed, the function returns the default locale, i.e.,
            standard English `('eng', None)`.

        :param strict: Indicate whether the string representation of a locale
            has to be strictly compliant with RFC 4646, or whether a Java
            style locale (character `_` instead of `-`) is accepted.


        :return: A tuple `(language_code, country_code)`, where the first code
            represents a ISO 639-3 alpha-3 code (or alpha-2 code), and the
            second code a ISO 3166-1 alpha-2 code.
        """
        if locale is None:
            raise ValueError("Undefined value 'locale'")

        match = REGEX_LOCALE.match(locale)
        if match is None:
            if strict:
                raise Locale.MalformedLocaleException(
                    f'The string "{locale}" doesn\'t represent a valid locale')

            match = REGEX_PERMISSIVE_LOCALE.match(locale)
            if match is None:
                raise Locale.MalformedLocaleException(
                    f'The string "{locale}" doesn\'t represent any forms of a valid locale')

        _, locale_language_code, locale_country_code, language_code = match.groups()

        return (locale_language_code, locale_country_code.upper()) if language_code is None \
            else (language_code, None)

    @staticmethod
    def compose_locale(language_code: str, country_code: str = None) -> str:
        """
        Return the string representation of the locale specified with a ISO
        639-3 alpha-3 code (or alpha-2 code), optionally followed by a dash
        character `-` and a ISO 3166-1 alpha-2 code


        :param language_code: A ISO 639-3 alpha-3 code (or alpha-2 code).

        :param country_code: A ISO 3166-1 alpha-2 code.


        :return: A string representing a locale.
        """
        return language_code if country_code is None \
            else f'{language_code}-{country_code}'


DEFAULT_LOCALE = Locale('eng')



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import re

from majormode.perseus.constant.contact import ContactName
from majormode.perseus.constant.privacy import Visibility
from majormode.perseus.constant.regex import REGEX_PATTERN_EMAIL_ADDRESS
from majormode.perseus.constant.regex import REGEX_PATTERN_PHONE_NUMBER
from majormode.perseus.utils import cast


class Contact:
    """
    Represent a contact information, such as an e-mail address, a phone
    numbers, the Uniform Resource Locator (URL) of a Web site.
    """
    class InvalidContactException(ValueError):
        """
        Indicate that a given name doesn't represent a valid contact name.
        """

    REGEX_EMAIL_ADDRESS = re.compile(REGEX_PATTERN_EMAIL_ADDRESS)
    REGEX_PHONE_NUMBER = re.compile(REGEX_PATTERN_PHONE_NUMBER)

    __CONTACT_NAME_REGEX_MAPPING = {
        ContactName.EMAIL: REGEX_EMAIL_ADDRESS,
        ContactName.PHONE: REGEX_PHONE_NUMBER,
    }

    def __init__(
            self,
            property_name: ContactName,
            property_value: str,
            is_primary: bool = False,
            is_verified: bool = False,
            strict: bool = True,
            visibility: Visibility = Visibility.public
    ):
        """
        Build a new instance `Contact`.


        :param property_name: The type of this contact information.

        :param property_value: The value of this contact, such as for instance
            `+84.8272170781`, the formatted value for a telephone number
            property.

        :param is_primary: Indicate whether this contact property is the first
            to be used to contact the entity that this contact information
            corresponds to.  There is only one primary contact property for a
            given property name (e.g., `EMAIL`, `PHONE`, `WEBSITE`).

        :param is_verified: Indicate whether this contact information has been
            verified, whether it has been grabbed from a trusted Social
            Networking Service (SNS), or whether through a challenge/response
            process.

        :param visibility: The visibility of this contact information to other
            users.  By default, `Visibility.public` if not defined.


        :raise AssertError: If the specified type of this contact information
            is not a string representation of an item of the enumeration
            `ContactName`.

        :raise ValueError: If the value of this contact information is null.
        """
        if isinstance(property_name, str):
            property_name = cast.string_to_enum(property_name, ContactName)
        else:
            if property_name not in ContactName:
                raise self.InvalidContactException(
                    f'The name "{property_name} doesn\'t represent a valid contact property name')

        if property_value is None:
            raise self.InvalidContactException("The property value of a contact MUST NOT be null")

        self.__property_name = property_name
        self.__property_value = property_value.strip().lower()

        if strict:
            self.assert_contact_value(self.__property_name, self.__property_value)

        self.__is_primary = is_primary
        self.__is_verified = is_verified
        self.__visibility = visibility or Visibility.public

    def __eq__(self, other):
        """
        Check whether this contact information to the one passed to this
        function.

        Two instances are equivalent when they are of the same type and they
        have the same value.


        :param other: an instance `Contact`.


        :return: `True` if this contact information corresponds to the
            contact information passed to this function; `False` otherwise.
        """
        return self.__property_name == other.__property_name \
            and self.__property_value == other.__property_value

    def __str__(self):
        """
        Return a nicely printable string representation of this contact
        information:

            [ name:string, value:string, [is_primary:boolean, [is_verified:boolean]] ]


        :return: a string of the array representation of this contact
            information.
        """
        # Do not include the attribute `is_verified` if the attribute
        # `is_primary` has not been defined.
        attributes = (
            str(self.__property_name),
            self.__property_value,
            self.__is_primary,
            self.__is_primary and self.__is_verified)

        return str([
            attribute
            for attribute in attributes
            if attribute is not None
        ])

    def assert_contact_value(
            self,
            property_name: ContactName,
            property_value: str
    ) -> None:
        """
        Check that the contact value is valid.

        :param property_name: The contact information's type.

        :param property_value: The contact's value, such as for instance
            `+84.8272170781`, the formatted value for a telephone number
            property.


        :raise InvalidContactException: If the value of the contact doesn't
            comply with the type of this contact.
        """
        if not self.__CONTACT_NAME_REGEX_MAPPING[property_name].match(property_value):
            raise self.InvalidContactException(
                f"Invalid value \"{property_value}\" of a contact information {property_name}"
            )

    @staticmethod
    def from_json(payload):
        """
        Convert the JSON expression of a contact information into an instance
        `Contact`.


        :param payload: a JSON expression representing a contact information:

            [ type:ContactName, value:string, [is_primary:boolean, [is_verified:boolean]] ]


        :return: an instance `Contact`.


        :raise AssertError: if the specified type of this contact information
            is not a string representation of an item of the enumeration
            `ContactName`.

        :raise ValueError: if the value of this contact information is null.
        """
        contact_item_number = len(payload)
        assert 2 <= contact_item_number <= 4, 'Invalid contact information format'

        is_primary = is_verified = None

        # Unpack the contact information to each component.
        if contact_item_number == 2:
            property_name, property_value = payload
        elif contact_item_number == 3:
            property_name, property_value, is_primary = payload
        else:
            property_name, property_value, is_primary, is_verified = payload

        return Contact(
            cast.string_to_enum(property_name, ContactName), property_value,
            is_primary=is_primary and cast.string_to_boolean(is_primary, strict=True),
            is_verified=is_verified and cast.string_to_boolean(is_verified, strict=True)
        )

    @staticmethod
    def from_object(obj):
        """
        Convert an object representing a contact information to an instance
        `Contact`.


        :param obj: an object containing the following attributes:

            * `name`: an item of the enumeration `ContactName` representing the
              type of this contact information.

            * `value`: value of this contact information representing by a string,
              such as `+84.01272170781`, the formatted value for a telephone
              number property.

            * `is_primary`: indicate whether this contact property is the first to
              be used to contact the entity that this contact information
              corresponds to.  There is only one primary contact property for a
              given property name (e.g., `EMAIL`, `PHONE`, `WEBSITE`).

            * `is_verified`: indicate whether this contact information has been
              verified, whether it has been grabbed from a trusted Social
              Networking Service (SNS), or whether through a challenge/response
              process.


        :raise ValueError: if the value of this contact information is null.
        """
        return obj if isinstance(obj, Contact) \
            else Contact(
                cast.string_to_enum(obj.property_name, ContactName),
                obj.property_value,
                is_primary=obj.is_primary and cast.string_to_boolean(obj.is_primary, strict=True),
                is_verified=obj.is_verified and cast.string_to_boolean(obj.is_verified, strict=True))

    @classmethod
    def from_string(
            cls,
            property_value: str,
            is_primary: bool = None,
            is_verified: bool = None,
            visibility: bool = None
    ) -> Contact:
        """
        Return the contact information corresponding to a contact value.


        :param property_value: A string representing a contact value.

        :param is_primary: Indicate whether this contact property is the first
            to be used to contact the entity that this contact information
            corresponds to.  There is only one primary contact property for a
            given property name (e.g., `EMAIL`, `PHONE`, `WEBSITE`).

        :param is_verified: Indicate whether this contact information has been
            verified, whether it has been grabbed from a trusted Social
            Networking Service (SNS), or whether through a challenge/response
            process.

        :param visibility: An item of the enumeration `Visibility` that
            indicates the visibility of this contact information to other
            users.  By default, `Visibility.public` if not defined.


        :return: An object `Contact` with the appropriate contact name and the
            given contact value.


        :raise InvalidContactException: If the contact value doesn't
            correspond to a valid contact information.
        """
        for property_name, regex in cls.__CONTACT_NAME_REGEX_MAPPING.items():
            if regex.match(property_value):
                return Contact(
                    property_name,
                    property_value,
                    is_primary=is_primary,
                    is_verified=is_verified,
                    strict=False,
                    visibility=visibility)
        else:
            raise cls.InvalidContactException(f'Unsupported contact information "{property_value}"')

    @property
    def is_primary(self):
        return self.__is_primary

    @is_primary.setter
    def is_primary(self, is_primary):
        if not isinstance(is_primary, bool):
            raise TypeError("argument 'is_primary' MUST be a boolean value")

        self.__is_primary = is_primary

    @property
    def is_verified(self):
        return self.__is_verified

    @property
    def property_name(self):
        return self.__property_name

    @property
    def property_value(self):
        return self.__property_value

    @property
    def visibility(self):
        return self.__visibility



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import re
import sys


REGEX_EGG_PACKAGE_PATH = re.compile(r'^.+[.]egg$')


def find_module_egg_path(module_path):
    """
    Find the path of the deployed egg package that may contain the
    specified module path.


    @param module_path: the path of a Python module.


    @return: the absolute path of the deployed egg package that contains
        the specified module path, or `None` if no deployed egg package
        contains this module path.
    """
    if module_path.find('.egg') == -1:
        return None

    module_absolute_path = os.path.abspath(module_path)

    egg_paths = [
        os.path.relpath(module_absolute_path, egg_path)
        for egg_path in [
            egg_path for egg_path in sys.path
            if REGEX_EGG_PACKAGE_PATH.match(egg_path) and module_absolute_path.startswith(egg_path)]]

    return None if len(egg_paths) == 0 else egg_paths[0]



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.model.enum import Enum


ApplicationMemberRole = Enum(
    # Secondary contact for teams and has many of the  same responsibilities
    # as the agent role. Administrators have access to all applications
    # registered on behalf of the team that these administrators belong to.
    'administrator',

    # Responsible for entering into legal agreements with the Service
    # Provider of the online platform where server applications are deployed
    # to or connect to.  The person who completes program enrollment is
    # assigned the agent role.
    'agent',

    # Manage development and delivery of an application.
    'developer',
)

ApplicationPlatform = Enum(
    # Command line interface (CLI) application that runs on a terminal console.
    'console',

    # Graphical User Interface (GUI) application that runs stand-alone in a
    # desktop or laptop computer.
    'desktop',

    # Graphical User Interface (GUI) application that runs in a smartphone.
    'mobile',

    # Application that waits for requests from other applications and
    # responds to them, thus providing a "service" upon their request.
    # For example, a web server, a RESTful API, are server applications.
    'server',
)

ApplicationStage = Enum(
    # A testing environment that enables the isolated execution of an
    # application for independent evaluation, monitoring or testing.
    'sandbox',

    # A production environment where the application is made available to
    # real users (equivalent to ``EnvironmentStage.prod``).
    'live',
)




# Copyright (C) 2021 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from majormode.perseus.model.enum import Enum


# Code name of the supported mobile device platforms.
DevicePlatform = Enum(
    'ios',
    'android'
)

NotificationMode = Enum(
    # Indicate that a notification message is delivered to the specified
    # recipients when the request for the transmission of information is
    # initiated by the receiver or client application, and then is
    # responded by the publisher or server platform.
    'pull',

    # Indicate that a notification message is delivered to the specified
    # recipients when the request for the transmission of information is
    # initiated by the publisher or server platform and pushed out to the
    # receiver or client application.
    'push'
)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import hashlib
import itertools
import time


def calculate_base_sequence(base_offsets):
    """
    Return a sequence of characters corresponding to the specified base.

    Example:

    ```python
    >>> calculate_base_sequence([ ('0', '9'), ('A', 'F') ])
    '0123456789ABCDEF'
    ```
    

    @param base_offsets: a list of character offsets in the form of tuples
        ``(x, y)`` where ``x`` corresponds to the first character of a
        sequence and ``y`` to the last character of this sequence.


    @return a sequence of characters, i.e., a string, corresponding to the
        specified base.
    """
    return ''.join([chr(c) for c in list(itertools.chain.from_iterable(
        [range(ord(x), ord(y) + 1) for x, y in base_offsets]))])


BASE62_OFFSETS = [('0', '9'), ('a', 'z'), ('A', 'Z')]
BASE62 = calculate_base_sequence(BASE62_OFFSETS)


def generate_secured_key(
        value,
        key_nonce_separator='.',
        nonce_length=4,
        base=BASE62):
    """
    Generate a secured key composed of the integer value encoded in Base62
    and a nonce.


    @param value: an integer value.

    @param key_nonce_separator: the character that is used to separate the
        key and the nonce to form the secured key.

    @param nonce_length: the number of characters to compose the nonce.

    @param base: a sequence of characters that is used to encode the
           integer value.


    @return: a tuple ``(key, nonce, secured_key)``:
        * ``key``: string representation of the integer value in Base62.

        * ``nonce``: "number used once", a pseudo-random number to
          ensure that the key cannot be reused in replay attacks.

        * ``secured_key``: a string composed of the key concatenated with
          the nonce.
    """
    if not isinstance(value, int):
        raise ValueError()

    posix_time = int(time.mktime(datetime.datetime.now().timetuple()))
    nonce = hashlib.md5(str(posix_time)).hexdigest()[:nonce_length]
    key = int_to_key(value, base=base)
    return key, nonce, f'{key}{key_nonce_separator}{nonce}'


def int_to_key(value, base=BASE62):
    """
    Convert the specified integer to a key using the given base.


    @param value: a positive integer.

    @param base: a sequence of characters that is used to encode the
        integer value.


    @return: a key expressed in the specified base.
    """
    def key_sequence_generator(value, base):
        """
        Generator for producing sequence of characters of a key providing an
        integer value and a base of characters for encoding, such as Base62
        for instance.


        @param value: a positive integer.

        @param base: a sequence of characters that is used to encode the
            integer value.


        @return: the next character of the object's key encoded with the
                 specified base.
        """
        base_length = len(base)
        while True:
            yield base[value % base_length]
            if value < base_length:
                break
            value //= base_length

    return ''.join([c for c in key_sequence_generator(value, base)])


def key_to_int(key, base=BASE62):
    """
    Convert the following key to an integer.


    @param key: a key.

    @param base: a sequence of characters that was used to encode the
        integer value.


    @return: the integer value corresponding to the given key.


    @raise ValueError: if one character of the specified key doesn't match
        any character of the specified base.
    """
    base_length = len(base)

    value = 0
    for c in reversed(key):
        value = (value * base_length) + base.index(c)

    return value


def parse_secured_key(
        secured_key,
        key_nonce_separator='.',
        nonce_length=4,
        base=BASE62):
    """
    Parse a given secured key and return its associated integer, the key
    itself, and the embedded nonce.


    @param secured_key a string representation of a secured key composed
        of a key in Base62, a separator character, and a nonce.

    @param key_nonce_separator: the character that is used to separate the
        key and the nonce to form the secured key.

    @param nonce_length: the number of characters to compose the nonce.

    @param base: a sequence of characters that is used to encode the
           integer value.


    @return: a tuple ``(value, key, nonce)``:

        * ``value``: the integer value of the key.

        * ``key``: the plain-text key.

        * ``nonce``: "number used once", a pseudo-random number to ensure that
          the key cannot be reused in replay attacks.


    @raise ValueError: if the format of the secured key is invalid, or if
        the embedded nonce is of the wrong length.
       """
    parts = secured_key.split(key_nonce_separator)
    if len(parts) != 2:
        raise ValueError("Invalid secured key format")

    key, nonce = parts
    if len(nonce) != nonce_length:
        raise ValueError("Invalid length of the key nonce")

    return key_to_int(key, base=base), key, nonce



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import functools
import re
import sys
import unidecode

from majormode.perseus.constant.regex import REGEX_PATTERN_EMAIL_ADDRESS
from majormode.perseus.constant.regex import REGEX_PATTERN_PHONE_NUMBER
from majormode.perseus.constant.regex import REGEX_PATTERN_USERNAME

REGEX_EMAIL_ADDRESS = re.compile(REGEX_PATTERN_EMAIL_ADDRESS)
REGEX_PHONE_NUMBER = re.compile(REGEX_PATTERN_PHONE_NUMBER)
REGEX_USERNAME = re.compile(REGEX_PATTERN_USERNAME)


def flatten_list(l):
    """
    Flatten the elements contained in the sub-lists of a list.


    :param l: A list containing sub-lists of elements.


    :return: A list with all the elements flattened from the sub-lists of
        the list `l`.
    """
    return [e for sublist in l for e in sublist]


def inicap_string(s, cleanse=False):
    """
    Convert the first letter of each word to capital letter without
    touching the rest of the word.


    :param s: a string.

    :param cleanse: `True` to remove any separator character from the
        string, such as comma; `False` to keeps any character but space.


    :return: a string for which the first letter of each word has been
        capitalized without modifying the case of the rest of the word.
    """
    return s and ' '.join([
        word[0].upper() + word[1:]
        for word in (s.split() if cleanse else s.split(' '))])


def is_email_address(s):
    """
    Indicate whether a string corresponds to an email address


    :param s: A string that is expected to be an email address.


    :return: `True` if the string corresponds to an email address,
        `False` otherwise.
    """
    return (s and REGEX_EMAIL_ADDRESS.match(s.strip().lower())) is not None


def is_phone_number(s):
    """
    Indicate whether a string corresponds to an international phone number

    The phone number MUST be written in E.164 numbering plan, formatted
    according to RFC 5733 (Extensible Provisioning Protocol (EPP) Contact
    Mapping).

    EPP-style phone numbers use the format `+CCC.NNNNNNNNNNxEEEE`, where
    `C` is the 1–3 digit country code, `N` is up to 14 digits, and `E` is
    the (optional) extension.  The leading plus sign and the dot following
    the country code are required.  The literal "x" character is required
    only if an extension is provided.

    :param s: An EPP-style phone number.


    :return: `True` if the string represents an international phone number,
        `False` otherwise.
    """
    return (s and REGEX_PHONE_NUMBER.match(s.strip())) is not None


def is_username(s):
    """
    Indicate whether a string corresponds to a username


    :param s: A username


    :return: `True` if the string represents a username, `False` otherwise.
    """
    return (s and REGEX_USERNAME.match(s.strip())) is not None


def string_hash_code(s):
    """
    Calculate the hash code of a string.


    :param s: a string.


    :return: the hash code of the string.
    """
    string_length = len(s)
    return functools.reduce(
        lambda x, y: x + y,
        [ord(s[i]) * 31 ** (string_length - i - 1) for i in range(string_length)]) % sys.maxsize


def strip_string_words(s):
    """
    Return the string with all its words stripped.


    :param s: A string.


    :return: The string with all its words stripped.
    """
    return s and ' '.join([word for word in s.split(' ')])


def string_to_keywords(s, keyword_minimal_length=1):
    """
    Split the specified string(s) in a list of ASCII keywords

    Remove any punctuation character from the specified list of keywords,
    remove any double or more space character and represent Unicode
    characters in ASCII.


    :param s: A word or a list of words to strip out any punctuation
        characters and to convert to ASCII characters.

    :param keyword_minimal_length: Minimal number of characters of the
        keywords to be returned.


    :return: A list of distinct keywords cleansed from any special Unicode
        accentuated character, punctuation character, and double space
        character.


    :raise ValueError: If the argument is null, or if it is not a string
        or a list of string.
    """
    if s is None:
        return None

    if isinstance(s, str):
        s = [s]
    elif isinstance(s, (list, set, tuple)):
        if any(not isinstance(_, str) for _ in s):
            raise ValueError('Each item of the list `s` MUST be a string')
    else:
        raise ValueError('The argument `s` MUST be a string or a list of strings')

    # Convert the list of strings to ASCII lower characters.
    ascii_strings = [
        unidecode.unidecode(_.strip()).lower()
        for _ in s
    ]

    # Replace from the list of strings any punctuation character with space.
    punctuationless_strings = [
        re.sub(r'[.,\\/#!$%^&*;:{}=\-_`~()<>"\']', ' ', _)
        for _ in ascii_strings
    ]

    # Decompose the string into distinct keywords.
    keywords = set(flatten_list([
        _.split()
        for _ in punctuationless_strings
    ]))

    # Filter out sub-keywords of less than the required number of characters.
    keywords = [
        keyword
        for keyword in keywords
        if len(keyword) >= keyword_minimal_length
    ]

    return keywords



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import collections

from PIL import Image
import exifread

from majormode.perseus.model.enum import Enum


# Exif tag which value corresponds to the exposure time, which is the
# effective length of time a camera’s shutter is open.  The total
# exposure is proportional to this exposure time, or duration of light
# reaching the film or image sensor.
EXIF_TAG_EXPOSURE_TIME = 0x829A

# Exif tag which value corresponds to the f-number, which is the
# ratio of the lens’s focal length to the diameter of the entrance
# pupil.  It is a dimensionless number that is a quantitative measure
# of lens speed.
EXIF_TAG_F_NUMBER = 0x829D

# Exif tag which value corresponds to the focal length, which is the
# actual focal length of the lens of the camera, which is a measure of
# how strongly the system converges or diverges light. For an optical
# system in air, it is the distance over which initially collimated
# rays are brought to a focus.  A system with a shorter focal length
# has greater optical power than one with a long focal length; that
# is, it bends the rays more strongly, bringing them to a focus in a
# shorter distance.
EXIF_TAG_FOCAL_LENGTH = 0x920A

# Exif tag which value corresponds to the ISO speed ratings, which is
# the measure of a digital imaging systems.  ISO speed ratings of a
# digital camera are based on the properties of the sensor and the
# image processing done in the camera, and are expressed in terms of
# the luminous exposure H (in lux seconds) arriving at the sensor.
EXIF_TAG_ISO_SPEED_RATINGS = 0x8827

# Exif tag which value corresponds to the orientation, which indicates
# the orientation of the camera relative to the captured scene.
#EXIF_TAG_ORIENTATION = 0x0112
EXIF_TAG_ORIENTATION = 'Image Orientation'


ColorComponentFilter = Enum(
    'Red',
    'Green',
    'Blue',
    'Grey'
)

CropAlignment = Enum(
    'left_or_top',
    'right_or_bottom',
    'center'
)

CropForm = Enum(
    'rectangle',
    'circle'
)

Filter = Enum(
    'NearestNeighbor', # nearest neighbor
    'Bilinear',        # linear interpolation in a 2x2 environment
    'Bicubic',         # cubic spline interpolation in a 4x4 environment
    'AntiAlias'        # high-quality downsampling filter
)

PIL_FILTER_MAP = {
    Filter.NearestNeighbor: Image.NEAREST,
    Filter.Bilinear: Image.BILINEAR,
    Filter.Bicubic: Image.BICUBIC,
    Filter.AntiAlias: Image.LANCZOS,
}

PictureExposure = Enum(
    'OverExposed',
    'UnderExposed',
    'NormallyExposed'
)


def convert_image_to_rgb_mode(image, fill_color=(255, 255, 255)):
    """
    Convert the specified image instance to RGB mode.


    @param image: a Python Library Image (PIL) instance to convert its
        pixel format to RGB, discarding the alpha channel.

    @param fill_color: color to be used to fill transparent pixels when
        discarding the alpha channel.  By default, the white color.


    @return: a Python Library Image instance with pixel format of RGB.
    """
    if image.mode not in ('RGBA', 'LA'):
        return image

    # In most cases simply discarding the alpha channel will give
    # undesirable result, because transparent pixels also have some
    # unpredictable colors. It is much better to fill transparent pixels
    # with a specified color.
    background_image = Image.new(image.mode[:-1], image.size, fill_color)
    background_image.paste(image, image.split()[-1])
    return background_image


def generate_multiple_pixel_resolutions(
        image,
        pixel_resolutions,
        filter=Filter.NearestNeighbor,
        does_crop=False,
        crop_alignment=CropAlignment.center,
        match_orientation=False):
    """
    Generate multiple resolution images of the given image.


    @param image: a Python Library Image (PIL) instance to generate
       multiple pixel resolutions from.

    @param pixel_resolutions: a list of tuples  ``(logical_size, width, height)``
        where:

        * ``logical_size``: string representation of the image size, such
          as, for instance, "thumbnail", "small", "medium", "large".

        * ``width``: positive integer corresponding to the number of pixel
          columns of the image.

        * ``height``: positive integer corresponding to the number of
          pixel rows.

    @param filter: indicate the filter to use when resizing the image.

    @param does_crop: indicate whether to crop each generated images.

    @param crop_alignment: if the image needs to be cropped, select which
        alignment to use when cropping.

    @param match_orientation: indicate whether the given canvas size
        should be inverted to match the orientation of the image.


    @return: an iterator, known as a generator, that returns a Python
             Library Image (PIL) instance each the generator is called.
    """
    for (logical_size, width, height) in \
            sorted(pixel_resolutions, key=lambda pixel_resolution: pixel_resolution[1], reverse=True):
        yield logical_size, resize_image(
            image,
            (width, height),
            filter=filter,
            does_crop=does_crop,
            crop_alignment=crop_alignment,
            match_orientation=match_orientation)


def get_exposure(image, filters=None):
    """
    Determine the exposure of a photo, which can be under-exposed,
    normally exposed or over-exposed.


    @param image: a Python Library Image (PIL) object to determine the
        exposure.

    @param filters: a list of ``ColorComponentFilter`` filter or ``None``
        to use all the filters.


    @return: an ``ExposureStatus`` instance that represents the exposure
             of the given PIL object.
    """
    def _get_exposure(histogram):
        total = sum(histogram)
        range_offset = len(histogram) / 4
        dark = float(sum(histogram[0:range_offset])) / total
        #normal = float(sum(histogram[range_offset:-range_offset])) / total
        light = float(sum(histogram[-range_offset:])) / total
        return PictureExposure.UnderExposed if dark > 0.5 and light < 0.5 \
            else PictureExposure.OverExposed if dark < 0.5 and light > 0.5 \
            else PictureExposure.NormallyExposed

    FILTER_SETTINGS = {
        ColorComponentFilter.Red: ('RGB', 0, 256),
        ColorComponentFilter.Green: ('RGB', 256, 512),
        ColorComponentFilter.Blue: ('RGB', 512, 768),
        ColorComponentFilter.Grey: ('L', 0, 256)
    }

    exposures = collections.defaultdict(int)
    for exposure in [ _get_exposure(image.convert(mode).histogram()[start_index:end_index])
        for (mode, start_index, end_index) in [ FILTER_SETTINGS[filtr]
            for filtr in filters or [
                ColorComponentFilter.Red,
                ColorComponentFilter.Green,
                ColorComponentFilter.Blue,
                ColorComponentFilter.Grey
            ] ] ]:
        exposures[exposure] += 1

    return sorted(exposures.iterkeys(), key=lambda k: exposures[k], reverse=True)[0]


def is_image_file_valid(file_path_name):
    """
    Indicate whether the specified image file is valid or not.


    @param file_path_name: absolute path and file name of an image.


    @return: ``True`` if the image file is valid, ``False`` if the file is
        truncated or does not correspond to a supported image.
    """
    # Image.verify is only implemented for PNG images, and it only verifies
    # the CRC checksum in the image.  The only way to check from within
    # Pillow is to load the image in a try/except and check the error.  If
    # as much info as possible is from the image is needed,
    # ``ImageFile.LOAD_TRUNCATED_IMAGES=True`` needs to bet set and it
    # will attempt to parse as much as possible.
    try:
        with Image.open(file_path_name) as image:
            image.load()
    except IOError:
        return False

    return True


def open_and_reorient_image(handle):
    """
    Load the image from the specified file and orient the image accordingly
    to the Exif tag that the file might embed, which would indicate the
    orientation of the camera relative to the captured scene.


    @param handle: a Python file object.


    @return: an instance returned by the Python Library Image library.
    """
    # Retrieve tags from the Exchangeable image file format (Exif)
    # included in the picture.  If the orientation of the picture is not
    # top left side, rotate it accordingly.

    # @deprecated
    # exif_tags = dict([ (exif_tag.tag, exif_tag)
    #     for exif_tag in exif.process_file(handle).itervalues()
    #         if hasattr(exif_tag, 'tag') ])

    exif_tags = exifread.process_file(handle)
    exif_tag_orientation = exif_tags.get(EXIF_TAG_ORIENTATION)

    rotation_angle = exif_tag_orientation and {
            3: 180,
            6: 270,
            8: 90
        }.get(exif_tag_orientation.values[0])

    handle.seek(0)  # exif.process_file has updated the file's current position.
    image = Image.open(handle)

    return image if rotation_angle is None else image.rotate(rotation_angle)
    # if rotation_angle:
    #     (width, height) = image.size
    #
    #     image = image.rotate(rotation_angle)
    #
    #     # @patch: PIL doesn't adjust the size of a rotated image!
    #     if rotation_angle in (90, 270) and width != height and image.size == (width, height):
    #         image.size = (height, width)
    #
    # return image


def realign_image(image, shift_x, shift_y, max_shift_x, max_shift_y,
        shift_angle=0.0, filter=Filter.NearestNeighbor,
        stretch_factor=None,
        cropping_box=None):
    """
    Realign the given image providing the specific horizontal and vertical
    shifts, and crop the image providing the maximum horizontal and
    vertical shifts of images of a same set so that they all have the same
    size.


    @param image: an instance of Python Image Library (PIL) image.

    @param shift_x: horizontal shift in pixels of the image.

    @param shift_y: vertical shift in pixels of the image.

    @param max_shift_x: maximum absolute value of the horizontal shift in
        pixels of the images in the capture.

    @param max_shift_y: maximum absolute value of the vertical shift in
        pixels of the images in the capture.

    @param shift_angle: horizontal displacement angle in degrees of
           the image.

    @param filter: indicate the filter to use when rotating the image.

    @param stretch_factor: coefficient of multiplication to stretch or
           to shrink the image in both horizontal and vertical
           directions.

    @param cropping_box: a 4-tuple defining the left, upper, right, and
           lower pixel coordinate of the rectangular region from the
           specified image.


    @return: a new instance of Python Image Library (PIL) image.
    """
    (width, height) = image.size

    # Determine the new size of the image based on the maximal horizontal
    # and vertical shifts of all the other images in this series.
    new_width = width - max_shift_x * 2
    new_height = height - max_shift_y * 2

    # Determine the coordinates of the zone to crop to center the image
    # based on the horizontal and vertical shifts.
    bounding_box_x = (0 if shift_x < 0 else shift_x * 2)
    bounding_box_y = (0 if shift_y < 0 else shift_y * 2)

    if max_shift_x > shift_x:
        bounding_box_width = width - abs(shift_x) * 2
        bounding_box_x += (bounding_box_width - new_width) / 2
        #bounding_box_x = max_shift_x - abs(shift_x)

    if max_shift_y > shift_y:
        bounding_box_height = height - abs(shift_y) * 2
        bounding_box_y += (bounding_box_height - new_height) / 2
        #bounding_box_y = max_shift_y - abs(shift_y)

    # Crop the image and rotate it based on the horizontal displacement
    # angle of this image.
    image = image.crop((bounding_box_x, bounding_box_y,
                        bounding_box_x + new_width, bounding_box_y + new_height)) \
        .rotate(-shift_angle, PIL_FILTER_MAP[filter])

    # Stretch or shrink this image based on its coefficient, keeping the
    # calculated new size of this image.
    if stretch_factor and stretch_factor > 0:
        image = image.resize((int(round(width * stretch_factor)), int(round(height * stretch_factor))),
                             PIL_FILTER_MAP[filter])
        (width, height) = image.size

        if stretch_factor >= 1:
            image = image.crop(((width - new_width) / 2, (height - new_height) / 2,
                                (width - new_width) / 2 + new_width - 1, (height - new_height) / 2 + new_height - 1))
        else:
            _image = Image.new(image.mode, (new_width, new_height))
            _image.paste(image, ((new_width - width) / 2, (new_height - height) / 2))
            image = _image

    # Crop the image to the specified rectangle, if any defined.
    if cropping_box:
        (crop_x1, crop_y1, crop_x2, crop_y2) = cropping_box
        (width, height) = image.size

        image = image.crop(
            (int(round(crop_x1 * width)), int(round(crop_y1 * height)),
             int(round(crop_x2 * width)), int(round(crop_y2 * height))))

    return image


def resize_image(image, canvas_size,
        filter=Filter.NearestNeighbor,
        does_crop=False,
        crop_alignment=CropAlignment.center,
        crop_form=CropForm.rectangle,
        match_orientation=False):
    """
    Resize the specified image to the required dimension.


    @param image: a Python Image Library (PIL) image instance.

    @param canvas_size: requested size in pixels, as a 2-tuple
        ``(width, height)``.

    @param filter: indicate the filter to use when resizing the image.

    @param does_crop: indicate whether the image needs to be cropped.

    @param crop_alignment: if the image needs to be cropped, select which
       alignment to use when cropping.

    @param match_orientation: indicate whether the given canvas size
        should be inverted to match the orientation of the image.


    @return: a PIL image instance corresponding to the image that has been
        resized, and possibly cropped.
    """
    (source_width, source_height) = image.size
    source_aspect = source_width / float(source_height)

    (canvas_width, canvas_height) = canvas_size
    canvas_aspect = canvas_width / float(canvas_height)

    if match_orientation:
        if (source_aspect > 1.0 > canvas_aspect) or (source_aspect < 1.0 < canvas_aspect):
            (canvas_width, canvas_height) = (canvas_height, canvas_width)
            canvas_aspect = canvas_width / float(canvas_height)

    if does_crop:
        if source_aspect > canvas_aspect:
            destination_width = int(source_height * canvas_aspect)
            offset = 0 if crop_alignment == CropAlignment.left_or_top \
                else source_width - destination_width if crop_alignment == CropAlignment.right_or_bottom \
                else (source_width - destination_width) / 2
            box = (offset, 0, offset + destination_width, source_height)
        else:
            destination_height = int(source_width / canvas_aspect)
            offset = 0 if crop_alignment == CropAlignment.left_or_top \
                else source_height - destination_height if crop_alignment == CropAlignment.right_or_bottom \
                else (source_height - destination_height) / 2
            box = (0, offset, source_width, destination_height + offset)

    else:
        if canvas_aspect > source_aspect:
            # The canvas aspect is greater than the image aspect when the canvas's
            # width is greater than the image's width, in which case we need to
            # crop the left and right edges of the image.
            destination_width = int(canvas_aspect * source_height)
            offset = (source_width - destination_width) / 2
            box = (offset, 0, source_width - offset, source_height)
        else:
            # The image aspect is greater than the canvas aspect when the image's
            # width is greater than the canvas's width, in which case we need to
            # crop the top and bottom edges of the image.
            destination_height = int(source_width / canvas_aspect)
            offset = (source_height - destination_height) / 2
            box = (0, offset, source_width, source_height - offset)

    return image.crop(box).resize((canvas_width, canvas_height), PIL_FILTER_MAP[filter])



# Copyright (C) 2020 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import os

from majormode.perseus.constant.stage import EnvironmentStage
from majormode.perseus.utils import cast


# List of the standard environment stages.
DEFAULT_ENVIRONMENT_STAGES = [
    EnvironmentStage.dev,
    EnvironmentStage.int,
    EnvironmentStage.stg,
    EnvironmentStage.prod,
]


def detect_environment_stage(
        root_path,
        environment_stages=None,
        default_stage=EnvironmentStage.dev):
    """
    Return the environment stage where this server software application
    has been deployed into.

    By convention, the deployment path of a server software application
    MUST respect the following file system hierarchy::

        /srv/<project>/<env>/<version>/


    @param root_path: Absolute path where the server software application
        is deployed in.

    @param environment_stages: A list of items of the enumeration
        `EnvironmentStage`.

    @param default_stage: An item of the enumeration `EnvironmentStage`
        corresponding to the default environment stage to return when the
        deployment path of the server software application doesn't
        respect the expected file system hierarchy.


    @return: An item of the enumeration `EnvironmentStage` referring to
        the environment stage where the server software application has
        been deployed into.
    """
    if environment_stages is None:
        environment_stages = DEFAULT_ENVIRONMENT_STAGES

    environment_stage_str = os.path.basename(os.path.dirname(os.path.realpath(os.path.join(root_path, '..'))))
    try:
        environment_stage = cast.string_to_enum(environment_stage_str, EnvironmentStage)
    except ValueError:
        environment_stage = cast.string_to_enum(
            input(f"Enter environment stage ({', '.join([str(s) for s in environment_stages])}): "),
            EnvironmentStage)

    return environment_stage



# Copyright (C) 2015 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import contextlib
import io
import logging
import os
import tempfile
import urllib.request
import zipfile


# Default number of bytes to be fetched at every read operation of the
# ZIP archive file data hosted on a computer network.
DEFAULT_CHUNK_SIZE = 8192


def download_file(
        url,
        chunk_size=DEFAULT_CHUNK_SIZE,
        file_path_name=None,
        memory_mapped=True):
    """
    Return a `ZipFile` instance of a ZIP archive file downloaded from
    the specified Uniform Resource Locator (URL).


    @param url: Uniform Resource Locator that specifies the location of
        the ZIP archive file on a computer network and a mechanism for
        retrieving it.

    @param chunk_size: Maximal number of bytes to be fetched every read
        operation of the ZIP archive file data hosted on the computer
        network.

    @param file_path_name: An absolute path and name of the file in
        which the archive will be stored in.  If not defined, the function
        downloads the archive file in a temporary file created in the most
        secure manner possible; the user is responsible for deleting this
        temporary file when done with it.

    @param memory_mapped: Indicate whether to map the ZIP archive file-
        like object into memory, for performance optimization, or whether
        to store this archive into disk, for memory optimization.
        If the archive is directly stored into disk, the caller is
        responsible for deleting the temporary file when done with it.


    @return: An object `ZipFile` of the remote ZIP archive file if the
        argument `memory_mapped` is set to `True`, or a tuple
        `(file_path_name, zip_file)`.


    @raise urllib.error.URLError: If a connection timed out occurs.
    """
    if memory_mapped:
        fd = io.StringIO()
    else:
        if not file_path_name:
            # Create a temporary file in the most secure manner possible.
            file_descriptor, file_path_name = tempfile.mkstemp()
            fd = os.fdopen(file_descriptor, 'wb')
        else:
            fd = open(file_path_name, 'wb')

    with contextlib.closing(urllib.request.urlopen(url)) as resource_handle:
        while True:
            chunk_data = resource_handle.read(chunk_size)
            if not chunk_data: break
            fd.write(chunk_data)
            logging.debug(f"Downloaded {len(chunk_data)} bytes")

    if memory_mapped:
        fd.seek(0)
    else:
        fd.close()

    zip_file = zipfile.ZipFile(fd if memory_mapped else file_path_name)

    return zip_file if memory_mapped else (zip_file, file_path_name)


def open_entry_file(zip_file, entry_file_name):
    """
    Extract the specified entry from the archive and return it as a file-
    like object.


    @param zip_file: An object `ZipFile`.

    @param entry_file_name: The name of the file in the archive, or a
        `ZipInfo` object.


    @return: A file-like object corresponding to the specified entry.


    @raise KeyError: If the specified entry name is not contained in the
        ZIP archive.
    """
    data = zip_file.read(entry_file_name)

    memory_mapped_file = io.StringIO()
    memory_mapped_file.write(data)
    memory_mapped_file.seek(0)

    return memory_mapped_file



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.model.enum import Enum


EnvironmentStage = Enum(
    # [Development]
    # This is the working environment for individual developers or small
    # teams.  Working in isolation with the rest of the tiers, the
    # developer(s) can try radical changes to the code without adversely
    # affecting the rest of the development team.
    'dev',

    # [Integration]
    # A common environment where all developers commit code changes.  The
    # goal of this environment is to combine and validate the work of the
    # entire project team so it can be tested before being promoted to the
    # Staging Environment.  It is possible for Development and Integration
    # to be the same environment (as in the case where the developer does
    # not use a local copy of the source code).
    'int',

    # [Staging]
    # The staging tier is a environment that is as identical to the
    # production environment as possible.  The purpose of the Staging
    # environment is to simulate as much of the Production environment as
    # possible.  The Staging environment can also double as a
    # Demonstration/Training environment.
    'stg',

    # [Production]
    # The production tier might include a single machine or a huge cluster
    # comprising many machines.
    'prod',
)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.locale import Locale


class Label:
    """
    A label corresponds to a humanly-readable textual content written in a
    given locale, English by default.
    """
    def __init__(self, content, locale=DEFAULT_LOCALE):
        """
        Build a new ``Label`` instance.


        :param content: humanly-readable textual content of the label.

        :param locale: a ``Locale`` instance defining the language of the
            textual content of this label, English by default.


        :raise ValueError: if the argument ``locale`` is ``None``.
        """
        if locale is None:
            raise ValueError('The locale of a label CANNOT be null')

        if not isinstance(locale, Locale):
            raise TypeError('The locale MUST be an instance of `Locale`')

        self.__content = content.strip()
        self.__locale = locale

    @staticmethod
    def from_json(payload) -> Label:
        """
        Build an object ``Label`` from JSON data.

        :param payload: A JSON representation of a localized text.

            ```json
            {
              "content": string,
              "label": locale
            }


        :return: An object ``Label``.
        """
        return payload and Label(
            payload['content'],
            locale=Locale.from_string(payload['locale'])
        )

    @property
    def content(self) -> str:
        return self.__content

    @property
    def locale(self) -> Locale:
        return self.__locale



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import codecs
import contextlib
import errno
import hashlib
import os
import platform
import shutil
import sys


def build_tree_path_name(file_name, directory_depth=8):
    """
    Return a pathname built of the specified number of sub-directories,
    and where each directory is named after the nth letter of the filename
    corresponding to the directory depth.

    Examples::

      >>> build_tree_path_name('foo.txt', 2, '/')
      'f/o/'
      >>> build_tree_path_name('0123456789abcdef')
      '0/1/2/3/4/5/6/7/'


    @param file_name: name of a file, with or without extension.

    @param directory_depth: number of sub-directories to be generated.


    @return: a file pathname.
    """
    file_name_without_extension, _file_extension_ = os.path.splitext(file_name)
    max_depth = min(directory_depth, len(file_name_without_extension))

    pathname = os.path.join(*[
        c
        for c in file_name_without_extension[:max_depth]
    ])

    return pathname


def build_tree_file_path_name(file_name, directory_depth=8):
    """
    Return a file pathname which pathname is built of the specified
    number of sub-directories, and where each directory is named after the
    nth letter of the filename corresponding to the directory depth.

    Examples::

    ```python
    >>> build_tree_file_path_name('foo.txt', 2, '/')
    'f/o/foo.txt'
    >>> build_tree_file_path_name('0123456789abcdef')
    '0/1/2/3/4/5/6/7/0123456789abcdef'
    ```


    @param file_name: name of a file, with or without extension.

    @param directory_depth: number of sub-directories to be generated.


    @return: a file pathname.
    """
    path_name = build_tree_path_name(file_name, directory_depth)
    return os.path.join(path_name, file_name)


def find_root_path(absolute_path, relative_path):
    """
    Return the root path of a path relative to an absolute path.


    @param absolute_path: an absolute path that is ended by the specified
        relative path.

    @param relative_path: a relative path that ends the specified absolute
        path.


    @return: the root path of the relative path.
    """
    _absolute_path = os.path.normpath(absolute_path)
    _relative_path = os.path.normpath(relative_path)

    index = _absolute_path.rfind(_relative_path)
    if index == -1 or len(_relative_path) + index < len(_absolute_path):
        raise ValueError('The relative path does not end the specified absolute path')

    return _absolute_path[:index]


def get_file_checksum(file_path_name, hash_algorithm_name='md5'):
    """
    Generate the MD5 checksum of the specified file.


    @param file_path_name: the absolute path and name of the file to
        generate its MD5 checksum.

    @param hash_algorithm_name: specify the hash algorithm to use.  Refer
        to ``hashlib.algorithms`` to get the names of the hash algorithms
        guaranteed to be supported by this module.


    @return: hash digest returned as a string of double length, containing
        only hexadecimal digits.  This may be used to exchange the value
        safely in email or other non-binary environments.


    @note: the file can not be entirely read to memory, but it needs to
        be read by chunks of memory that will be freed after each
        iteration.  What's important to notice is that the file  must be
        opened in binary mode.  The function breaks the file into chunks
        using block size of any multiple of 128 (say 8192, 32768, etc.)
        and its feed them to MD5 consecutively using ``update()``.  This
        takes advantage advantage of the fact that MD5 has 128-byte digest
        blocks.  The function actually uses a block size that depends on the block
        size of the filesystem to avoid performances issues.

    @note: The ``F_FRSIZE`` value is the actual minimum allocation unit
        of the filesystem, while the ``F_BSIZE`` is the block size that
        would lead to most efficient use of the disk with io calls.
    """
    hash_algorithm = hashlib.new(hash_algorithm_name)

    if sys.platform == "win32":
        import ctypes

        sectors_per_cluster = ctypes.c_ulonglong(0)
        bytes_per_sector = ctypes.c_ulonglong(0)
        root_path_name = ctypes.c_wchar_p(u"C:\\")

        ctypes.windll.kernel32.GetDiskFreeSpaceW(
            root_path_name,
            ctypes.pointer(sectors_per_cluster),
            ctypes.pointer(bytes_per_sector),
            None,
            None)

        block_size = bytes_per_sector.value
    else:
        block_size = os.statvfs('/').f_bsize

    with open(file_path_name, 'rb') as handle:
        for chunk in iter(lambda: handle.read(block_size), b''):
            hash_algorithm.update(chunk)

    return hash_algorithm.hexdigest()


def make_directory_if_not_exists(path):
    """
    Create the specified path, making all intermediate-level directories
    needed to contain the leaf directory.  Ignore any error that would
    occur if the leaf directory already exists.


    @note: all the intermediate-level directories are created with the
        default mode is 0777 (octal).


    @param path: the path to create.


    @raise OSError: an error that would occur if the path cannot be
        created.
    """
    try:
        os.makedirs(path)
    except OSError as error:  # Ignore if the directory has been already created.
        if error.errno != errno.EEXIST:
            raise error


def move_file(source_file_pathname, destination_file_pathname):
    """
    Move the the specified file to another location.  If the destination
    already exists, it is replaced silently.

    This function is an alternative to ``shutil.move(src, dst)``, which
    might raise ``OSError`` if the destination already exists.


    @param source_file_pathname: the complete path and name of the file to
        move.

    @param destination_file_pathname: the complete path and name of the
        file once moved.
    """
    if os.path.exists(destination_file_pathname):
        os.remove(destination_file_pathname)

    shutil.move(source_file_pathname, destination_file_pathname)


def remove_file_if_exists(file_pathname):
    """
    Remove the file path, if it exists.


    @param file_pathname: path and file name the refers to the file.
    """
    try:
        os.remove(file_pathname)
    except:
        pass


@contextlib.contextmanager
def smart_open(file_path_name, mode='r', encoding='utf-8'):
    """
    Open an encoded file using the given mode and return a wrapped version
    providing transparent encoding/decoding.  The default file mode is 'r'
    meaning to open the file in read mode.


    @param file_path_name: the absolute path and name of the file to
        open.

    @param mode: indicate how the file is to be opened.  The most commonly-
        used values of mode are 'r' for reading, 'w' for writing
        (truncating the file if it already exists), and 'a' for appending
        (which on some Unix systems means that all writes append to the
        end of the file regardless of the current seek position).  If mode
        is omitted, it defaults to 'r'.

    @param encoding: specify the encoding which is to be used for the
        file.


    @return: an object of the file type.
    """
    if file_path_name and file_path_name != '-':
        file_handle = codecs.open(file_path_name, mode=mode, encoding='utf-8')
    else:
        file_handle = sys.stdout

    try:
        yield file_handle
    finally:
        if file_handle is not sys.stdout:
            file_handle.close()


def which(exe_name):
    """
    Locate a program file in the user's path.


    @param exe_name: name of the executable file.


    @return: `None` if the executable has not been found in the user's
        path, or the path for the executable file.
    """
    def is_exe(file_path_name):
        return os.path.isfile(file_path_name) and os.access(file_path_name, os.X_OK)

    is_platform_windows = (platform.system() == 'Windows')

    file_path, _file_name = os.path.split(exe_name)
    if file_path:
        if is_exe(exe_name):
            return exe_name
    else:
        for path in os.environ['PATH'].split(os.pathsep):
            exe_file_path_name = os.path.join(path, exe_name)
            if is_exe(exe_file_path_name):
                return exe_file_path_name

            if is_platform_windows:
                windows_exe_file_path_name = '%s.exe' % exe_file_path_name
                if is_exe(windows_exe_file_path_name):
                    return windows_exe_file_path_name

                windows_com_file_path_name = '%s.exe' % exe_file_path_name
                if is_exe(windows_com_file_path_name):
                    return windows_com_file_path_name

    return None



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# Regular expression to match an email address compliant with RFC 5322
# (http://tools.ietf.org/html/rfc5322#section-3.4.1).  This regular
# expression has two parts: the part before the `@`, and the part
# after the `@`.  There are two alternatives for the part before the
# `@`: it can either consist of a series of letters, digits and
# certain symbols, including one or more dots.  However, dots may not
# appear consecutively or at the start or end of the email address.
# The other alternative requires the part before the `@` to be
# enclosed in double quotes, allowing any string of ASCII characters
# between the quotes.  Whitespace characters, double quotes and
# backslashes must be escaped with backslashes.
#
# The part after the `@` also has two alternatives.  It can either
# be a fully qualified domain name (e.g., `somehost.example.com`).
# The literal Internet address can either be an IP address, or a
# domain-specific routing address.  This regular expression omits the
# syntax using double quotes and square brackets, as RFC 2822 marks
# this notation as obsolete.
#REGEX_PATTERN_EMAIL_ADDRESS = r"^[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$"
REGEX_PATTERN_EMAIL_ADDRESS = r"""(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])"""

# Regular expression that matches an integer.
REGEX_PATTERN_INTEGER = r'\d+'

# Regular expression that matches an Internet Protocol version 4 (32-bit
# number).
REGEX_PATTERN_IPV4 = r'(\d{1,3}).(\d{1,3}).(\d{1,3}).(\d{1,3})'

# Regular expression that matches a Media Access Control (MAC) address.
#
# @note: This regular expression is generated with the following code:
#
#     ```python
#     r'^%s$' % '[.:-]{0,1}'.join([r'([0-9A-Za-z]{2})' for i in range(6)])
#     ```
REGEX_PATTERN_MAC_ADDRESS = r'([0-9A-Za-z]{2})[.:-]{0,1}([0-9A-Za-z]{2})[.:-]{0,1}([0-9A-Za-z]{2})[.:-]{0,1}([0-9A-Za-z]{2})[.:-]{0,1}([0-9A-Za-z]{2})[.:-]{0,1}([0-9A-Za-z]{2})'

# Regular expression that matches a natural number including zero.
REGEX_PATTERN_NATURAL_NUMBER = r'^[0-9]+$'

# Regular expression to require a password to contain:
#
# - At least eight characters
# - At least one uppercase letter
# - At least one lowercase letter
# - At least one digit
# - At least one special character
REGEX_PATTERN_PASSWORD = r'^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*[@$!%*?&])[A-Za-z\d@$!%*?&]{8,}$'

# Regular expression that matches a phone number in E.164 numbering plan,
# formatted according to RFC 5733 (Extensible Provisioning Protocol (EPP)
# Contact Mapping).  EPP-style phone numbers use the format
# `+CCC.NNNNNNNNNNxEEEE`, where `C` is the 1–3 digit country code,
# `N` is up to 14 digits, and `E` is the (optional) extension.  The
# leading plus sign and the dot following the country code are required.
# The literal "x" character is required only if an extension is provided.
#
# References :
#
# - E.164 Numbering plan; https://www.itu.int/rec/T-REC-E.164-201011-I/en,
#   https://en.wikipedia.org/wiki/E.164
#
# - E.123 Notation for national and international telephone numbers,
#   e-mail addresses and web addresses; https://www.itu.int/rec/T-REC-E.123-200102-I/en,
#   https://en.wikipedia.org/wiki/E.123
#
#   For digit grouping, E.123 specifically recommends that:
#
#   - Only spaces be used to visually separate groups of numbers "unless
#     an agreed upon explicit symbol (e.g. hyphen) is necessary for
#     procedural purposes" in national notation;
#   - Only spaces should be used to visually separate groups of numbers
#     in international notation;
#   - Spaces should separate country code, area code and local number.
#
# - Extensible Provisioning Protocol (EPP) Contact Mapping
#   https://tools.ietf.org/html/rfc5733#section-2.5
#
#
# @patch: [2023-06-27] We made the dot (following the country code)
#   optional to support E.123 Notation for international telephone
#   numbers.
#
# @deprecated: REGEX_PATTERN_PHONE_NUMBER = r'(([+][(]?[0-9]{1,3}[)]?)|([(]?[0-9]{4}[)]?))\s*[)]?[-\s\.]?[(]?[0-9]{1,3}[)]?([-\s\.]?[0-9]{3})([-\s\.]?[0-9]{3,4})'
REGEX_PATTERN_PHONE_NUMBER = r'^\+[0-9]{1,3}\.{0,1}[0-9]{4,14}(?:x.+)?$'

# Pattern of a regular expression to match a username:
#
# 1. MUST only contain lowercase/uppercase letters, digits, underscores,
#    and periods
#
# 2. MUST start with a lowercase/uppercase letter
#
# 3. MUST NOT end with an underscore or a period
#
# 4. MUST NOT contain underscore and period next to each other (e.g.,
#    "user_.name")
#
# 5. MUST NOT contain underscore or period used multiple times in a row
#    (e.g., "user__name, "user..name")
#
# 6. MUST contain between 6 and 20 characters
REGEX_PATTERN_USERNAME = r'^(?=.{6,20}$)(?![0-9_.])(?!.*[_.]{2})[a-zA-Z0-9._]+(?<![_.])$'

# Regular expression that matches a Universally Unique Identifier (UUID).
REGEX_PATTERN_UUID = r'(?i)[\da-z]{8}-{0,1}[\da-z]{4}-{0,1}[\da-z]{4}-{0,1}[\da-z]{4}-{0,1}[\da-z]{12}'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import math

from majormode.perseus.model.obj import Serializable


class BoundingBox:
    """
    Represent a rectangle in geographical coordinates.
    """
    def __init__(self, southwest, northeast):
        """
        Build an instance `BoundinxBox` from the points at its south-west
        and north-east corners.


        :param southwest: an instance `GeoPoint` corresponding to the south-
            west corner of the bounding box.

        :param northeast: an instance `GeoPoint` corresponding to the north-
            east corner of the bounding box.
        """
        self.__southwest = southwest
        self.__northeast = northeast

    @property
    def northeast(self):
        return self.__northeast

    @property
    def southwest(self):
        return self.__southwest


class GeoPoint:
    """
    Represent the geographic coordinates of a location as reported by a
    location provider, such as Global Positioning System (GPS) or an
    Hybrid Positioning System based on wireless network.

    For more information about a variety of calculations for latitude/
    longitude points, with the formula and code fragments for implementing
    them, we recommend the following link:

    * Calculate distance, bearing and more between Latitude/Longitude
      points; Movable Type Scripts;
      http://www.movable-type.co.uk/scripts/latlong.html
    """

    # Code names of the supported location providers.
    LOCATION_PROVIDER_NAMES = ['gps', 'network', 'fused']

    # Radius of the Earth expressed in meters.
    EARTH_RADIUS_METERS = 6371009

    def __eq__(self, other):
        """
        Indicate whether this location corresponds to the specified other
        location.


        :param other: an other instance `GeoPoint`.


        :return: `True` if these two instance represents the same geographic
            location, at least the same longitude and latitude, and the same
            altitude if defined in both instances.
        """
        if self.__longitude != other.__longitude or self.__latitude != other.__latitude:
            return False

        return self.__altitude is None or other.__altitude is None or \
            self.__altitude == other.__altitude

    def __hash__(self):
        return hash(str(self))

    def __init__(
            self,
            latitude,
            longitude,
            altitude=None,
            accuracy=None,
            bearing=None,
            fix_time=None,
            provider=None,
            speed=None):
        """
        Build a `GeoPoint` instance.


        :param latitude: Latitude-angular distance, expressed in decimal
            degrees (WGS84 datum), measured from the center of the Earth, of 
            the point north or south of the Equator.

        :param longitude: Longitude-angular distance, expressed in decimal
            degrees (WGS84 datum), measured from the center of the Earth, of 
            the point east or west of the Prime Meridian.

        :param altitude: Altitude in meters of the point.

        :param accuracy: Accuracy in meters of the point.

        :param bearing: Bearing in degrees.  Bearing is the horizontal
            direction of travel of the device.  It is not related to the
            device orientation.  It is guaranteed to be in the range
            `[0.0, 360.0]`, or `null` if this device does not have a
            bearing.

        :param fix_time: Time when the fix of the location has been
            calculated.

        :param provider: Code name of the location provider that reported the
            geographical location:

            * `gps`: Indicate that the location has been provided by a
              Global Positioning System (GPS).

            * `network`: Indicate that the location has been provided by an
              hybrid positioning system, which uses different positioning
              technologies, such as Global Positioning System (GPS), Wi-Fi
              hotspots, cell tower signals.

        :param speed: Speed in meters/second over the ground, or `null` if
            this location does not have a speed.


        :raise ValueError: If some arguments are missing or of a wrong format
            or value.
        """
        if latitude is None or longitude is None:
            raise ValueError("Undefined latitude or longitude")

        if provider is not None:
            provider = provider.strip().lower()
            if provider not in GeoPoint.LOCATION_PROVIDER_NAMES:
                raise ValueError(f'Unsupported location provider "{provider}"')

        self.__latitude = float(latitude)
        self.__longitude = float(longitude)
        self.__altitude = altitude and float(altitude)
        self.__accuracy = accuracy and float(accuracy)
        self.__bearing = bearing and float(bearing)
        self.__speed = speed and float(speed)
        self.__fix_time = fix_time
        self.__provider = provider

    def __repr__(self):
        optional_arguments = (
            ('altitude', self.__altitude),
            ('accuracy', self.__accuracy),
            ('bearing', self.__bearing),
            ('fix_time', self.__fix_time and f"'{str(self.__fix_time)}'"),
            ('provider', self.__provider and f"'{self.__provider}'"),
            ('speed', self.__speed)
        )

        optional_arguments_str = ', '.join([
            f'{name}={value}'
            for name, value in optional_arguments
            if value is not None
        ])

        return f'{self.__class__.__name__}({self.__latitude}, {self.__longitude}' \
            f', {optional_arguments_str}' if optional_arguments_str else '' \
            ')'

    def __str__(self):
        return f'({self.__latitude}, {self.__longitude})' if not self.__altitude \
            else f'({self.__latitude}, {self.__longitude}, {self.__altitude})'

    @staticmethod
    def objectify_attributes(o):
        """
        Replace in place the geolocation attributes of an object to an
        attribute `GeoPoint` of this object grouping all these attributes.


        :param o: An object.  This object MUST at least contains the
            attributes `latitude` and `longitude`.
        """
        if o.longitude is None or o.latitude is None:
            return

        o.location = GeoPoint(
            o.latitude,
            o.longitude,
            altitude=getattr(o, 'altitude', None),
            accuracy=getattr(o, 'accuracy', None),
            bearing=getattr(o, 'bearing', None),
            fix_time=getattr(o, 'fix_time', None),
            provider=getattr(o, 'provider', None),
            speed=getattr(o, 'speed', None))

        del o.latitude
        del o.longitude
        if hasattr(o, 'altitude'):
            del o.altitude
        if hasattr(o, 'accuracy'):
            del o.accuracy
        if hasattr(o, 'bearing'):
            del o.bearing
        if hasattr(o, 'fix_time'):
            del o.fix_time
        if hasattr(o, 'provider'):
            del o.provider
        if hasattr(o, 'speed'):
            del o.speed

    def equirectangular_distance(self, other):
        """
        Return the approximate equirectangular when the location is close to
        the center of the cluster.

        For small distances, Pythagoras’ theorem can be used on an
        equirectangular projection.

        Equirectangular formula::

            x = Δλ ⋅ cos φm
            y = Δφ
            d = R ⋅ √(x² + y)²

        It will always over-estimate compared to the real Haversine distance.
        For example it will add no more than 0.05382 % to the real distance if
        the delta latitude or longitude between your two points does not
        exceed 4 decimal degrees.

        The standard formula (Haversine) is the exact one (that is, it works
        for any couple of longitude/latitude on earth) but is much slower as
        it needs 7 trigonometric and 2 square roots.  If your couple of points
        are not too far apart, and absolute precision is not paramount, you
        can use this approximate version (Equirectangular), which is much
        faster as it uses only one trigonometric and one square root:

        ```python
        Python 2.7.6rc1 (v2.7.6rc1:4913d0e9be30+, Oct 27 2013, 20:52:11)
        [GCC 4.2.1 (Apple Inc. build 5666) (dot 3)] on darwin
        Type "help", "copyright", "credits" or "license" for more information.
        >>> from majormode.perseus.model.geolocation import GeoPoint
        >>> import time
        >>>
        >>> source = GeoPoint(106.739036, 10.797977)
        >>> destination = GeoPoint(106.743325, 10.800195)
        >>>
        >>> start_time = time.time()
        >>> for i in xrange(1000000):
        ...     d = source.great_circle_distance(destination)
        ...
        >>> print time.time() - start_time
        5.62202811241
        >>> print d
        529.424701041
        >>>
        >>> start_time = time.time()
        >>> for i in xrange(1000000):
        ...     d = source.equirectangular_distance(destination)
        ...
        >>> print time.time() - start_time
        2.78262710571
        >>> print d
        529.424701073
            >>>
        ```


        :param other: a `GeoPoint` instance.


        :return: the great-circle distance, in meters, between this geographic
            coordinates to the specified other point.
        """
        x = math.radians(other.__longitude - self.__longitude) \
            * math.cos(math.radians(other.__latitude + self.__latitude) / 2)
        y = math.radians(other.__latitude - self.__latitude)

        return math.sqrt(x * x + y * y) * GeoPoint.EARTH_RADIUS_METERS

    def great_circle_distance(self, other):
        """
        Return the great-circle distance, in meters, from this geographic
        coordinates to the specified other point, i.e., the shortest distance
        over the earth’s surface, ‘as-the-crow-flies’ distance between the
        points, ignoring any natural elevations of the ground.

        Haversine formula::

          R = earth’s radius (mean radius = 6,371km)
          Δlat = lat2 − lat1
          Δlong = long2 − long1
          a = sin²(Δlat / 2) + cos(lat1).cos(lat2).sin²(Δlong/2)
          c = 2.atan2(√a, √(1−a))
          d = R.c


        :param other: a `GeoPoint` instance.


        :return: the great-circle distance, in meters, between this geographic
            coordinates to the specified other point.
        """
        distance_latitude = math.radians(abs(self.__latitude - other.__latitude))
        distance_longitude = math.radians(abs(self.__longitude - other.__longitude))

        a = math.sin(distance_latitude / 2) * math.sin(distance_latitude / 2) + \
            math.cos(math.radians(self.__latitude)) \
            * math.cos(math.radians(other.__latitude)) \
            * math.sin(distance_longitude / 2) \
            * math.sin(distance_longitude / 2)

        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

        return GeoPoint.EARTH_RADIUS_METERS * c

    @staticmethod
    def from_json(payload):
        """
        Build a `GeoPoint` instance from the specified JSON object.


        :param payload: JSON representation of a geographic location::

                {
                  "accuracy": decimal,
                  "altitude": decimal,
                  "bearing": decimal,
                  "longitude": decimal,
                  "latitude": decimal,
                  "provider": string,
                  "speed": decimal
                }

            where:

            * `accuracy`: accuracy of the position in meters.

            * `altitude`: altitude in meters of the location.

            * `bearing`: bearing in degrees.  Bearing is the horizontal direction
              of travel of the device, and is not related to the device orientation.
              It is guaranteed to be in the range `[0.0, 360.0]`, or `null` if
              this device does not have a bearing.

            * `latitude`: latitude-angular distance, expressed in decimal degrees
              (WGS84 datum), measured from the center of the Earth, of a point north
              or south of the Equator.

            * `longitude`: longitude-angular distance, expressed in decimal
              degrees (WGS84 datum), measured from the center of the Earth, of a
              point east or west of the Prime Meridian.

            * `provider`: code name of the location provider that reported the
              geographical location:

              * `gps`: indicate that the location has been provided by a Global
                Positioning System (GPS).

              * `network`: indicate that the location has been provided by an hybrid
                positioning system, which uses different positioning technologies,
                such as Global Positioning System (GPS), Wi-Fi hotspots, cell tower
                signals.

            * `speed`: speed in meters/second over the ground, or `None` if this
              location does not have a speed.


        :return: a `GeoPoint` instance or `None` if the JSON payload is nil.
        """
        return payload and GeoPoint(
            payload['latitude'],
            payload['longitude'],
            accuracy=payload.get('accuracy'),
            altitude=payload.get('altitude'),
            bearing=payload.get('bearing'),
            fix_time=payload.get('fix_time'),
            provider=payload.get('provider'),
            speed=payload.get('speed'))

    @staticmethod
    def convert_dd_to_dms_string(dd):
        """
        Convert decimal degrees (DD), which expresses latitude and longitude
        geographic coordinates as decimal fraction, to a degrees, minutes, and
        seconds (DMS) string representation, such as 38° 53' 23" N.


        :param dd: decimal degrees of the geographic coordinates of a location
            on the Earth.


        :return: degrees, minutes, and seconds (DMS) string representation of
            this location.
        """
        degrees = int(dd)
        minutes = (dd - degrees) * 60
        seconds = int((minutes - int(minutes)) * 60)
        return f"{degrees}° {int(minutes)}' {int(seconds)}\""

    @staticmethod
    def convert_dms_to_dd(
            degree_num, degree_den,
            minute_num, minute_den,
            second_num, second_den):
        """
        Convert the degree/minute/Second formatted GPS data to decimal
        degrees.


        :param degree_num: the numerator of the degree object.

        :param degree_den: the denominator of the degree object.

        :param minute_num: the numerator of the minute object.

        :param minute_den: the denominator of the minute object.

        :param second_num: the numerator of the second object.

        :param second_den: the denominator of the second object.


        :return: a decimal degree.
        """
        degree = float(degree_num) / float(degree_den)
        minute = float(minute_num) / float(minute_den) / 60
        second = float(second_num) / float(second_den) / 3600
        return degree + minute + second

    @staticmethod
    def convert_dms_string_to_dd(dms):
        """
        Convert a degrees, minutes, and seconds (DMS) string representation,
        such as 38° 53' 23" N, to a decimal degrees (DD), which expresses
        latitude and longitude geographic coordinates as decimal fraction.


        :param dms: degrees, minutes, and seconds (DMS) string representation
            of a location on the Earth.


        :return: decimal degrees of the geographic coordinates of a location.
        """
        degree_mark_offset = dms.find(u'°')
        degrees = float(dms[:degree_mark_offset].strip())
        minute_mark_offset = dms.find(u"'")
        minutes = float(dms[degree_mark_offset + 1:minute_mark_offset].strip())
        second_mark_offset = dms.find(u'"')
        seconds = float(dms[minute_mark_offset + 1:second_mark_offset].strip())
        return degrees + (minutes / 60) + (seconds / 3600)

    @staticmethod
    def convert_meter_to_dd(distance, latitude):
        """
        Convert distances expressed in meters to degrees.  A degree of
        longitude at the equator is 111.2 kilometers.  A minute is 1853
        meters.  A second is 30.9 meters.  For other latitudes multiply by
        cos(lat).  Distances for degrees, minutes and seconds in latitude
        are very similar and differ very slightly with latitude.


        :param distance: distance expressed in meters.

        :param latitude: latitude, expressed in degrees, of a point at the
            surface of the Earth.


        :return: approximation of the conversion of this distance in
            degrees for the specified latitude.
        """
        return distance * math.cos(math.radians(latitude)) / 111200

    @property
    def accuracy(self):
        return self.__accuracy

    @property
    def altitude(self):
        return self.__altitude

    @property
    def bearing(self):
        return self.__bearing

    @property
    def fix_time(self):
        return self.__fix_time

    @property
    def latitude(self):
        return self.__latitude

    @property
    def longitude(self):
        return self.__longitude

    @property
    def provider(self):
        return self.__provider

    @property
    def speed(self):
        return self.__speed


class WeightedGeoPointCluster:
    """
    Cluster of geographical points weighted with their accuracy.
    """
    def __calculate_center(self):
        """
        Determine the center of the cluster depending on the current list of
        geographical points and their respective accuracy .

        The more accurate a geographical point is, the more the center of the
        cluster moves toward this point.


        :return: An object `GeoPoint` to be used as the center of this cluster.
        """
        points = [
            (point.latitude, point.longitude, point.accuracy)
            for point in self.__points
        ]

        latitudes, longitudes, accuracies = list(zip(*points))
        coefficients = [1 / accuracy for accuracy in accuracies]

        coefficients_sum = sum(coefficients)

        center_latitude = sum([latitudes[i] * coefficients[i] for i in range(len(latitudes))]) / coefficients_sum
        center_longitude = sum([longitudes[i] * coefficients[i] for i in range(len(longitudes))]) / coefficients_sum
        center_accuracy = len(coefficients) / coefficients_sum

        return GeoPoint(center_latitude, center_longitude, accuracy=center_accuracy)

    def __init__(
            self,
            points=None):
        """
        Build a cluster of geographical points.


        :param points: A list of objects `GeoPoints`.
        """
        self.__points = points or []
        self.__is_dirty = bool(points)  # points is None or len(points) == 0
        self.__center = None

    def add(self, point):
        """
        Add a geographical point to the cluster.


        :param point: An object `GeoPoint`.


        :return: This object `WeightedGeoPointCluster`.
        """
        if not isinstance(point, GeoPoint):
            raise ValueError("The argument `point` must be an object `GeoPoint`")

        self.__points.append(point)
        self.__is_dirty = True
        return self

    @property
    def center(self):
        """
        Return the center of this cluster.


        :return: An object `GeoPoint` corresponding to the center of this
            cluster.
        """
        if self.__is_dirty:
            self.__center = self.__calculate_center()
            self.__is_dirty = False

        return self.__center



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from functools import reduce
from itertools import starmap
import io
import os
import re

from six import string_types


class Version:
    """
    The software component version numbering is mostly inspired from
    Apache version numbering or similar other version numbering such as
    Semantic Versioning.  Version numbers are denoted using a standard
    tuple of integers: ``major.minor.patch``.

    A ``major`` version identifies the product stage of the project.  The
    basic intent is that ``major`` versions are incompatible, large-scale
    upgrades of the software component.  This enables a check of a client
    application against the latest version of the software component to
    ensure compatibility.  If there is a discrepancy between the two, the
    client application MUST be updated accordingly.

    A ``minor`` version is incremented when substantial new functionality
    or improvement are introduced; the ``major`` version number doesn't
    change.  A ``minor`` version retains backward compatibility with older
    minor versions.  It is NOT forward compatible as a previous ``minor``
    version doesn't include new functionality or improvement that has been
    introduced in this newer ``minor`` version.

    A ``patch`` version is incremented when bugs were fixed or
    implementation details were refactored.  The ``major`` and ``minor``
    version don't change.  A ``patch`` version is backward and forward
    compatible with older and newer patches of the same major and minor
    version.

    References:

    * Apache Portable Runtime (APR)'s Version Numbering
      (http://apr.apache.org/versioning.html)
    * Semantic Versioning (http://semver.org/)
    * PEP 386 - Changing the version comparison module in Distutils
      (http://www.python.org/dev/peps/pep-0386/)
    """
    # Name of the file in which the version of an application is commonly
    # written.
    DEFAULT_VERSION_FILE_NAME = 'VERSION'

    # Regular expression that matches the string representation of a version
    # denoted using a standard tuple of integers ``major.minor.patch``.
    # REGEX_VERSION = re.compile(r'^(\d+)(.(\d+)(.(\d+)){0,1}){0,1}$')
    REGEX_PATTERN_SEMANTIC_VERSION = r'^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$'

    def __eq__(self, other: Version) -> int:
        result = self.compare_versions(
            (self.__major, self.__minor, self.__patch),
            (other.major, other.minor, other.patch)
        )
        return result == 0

    def __ge__(self, other) -> bool:
        return self.__eq__(other) or self.__gt__(other)

    def __gt__(self, other) -> bool:
        result = self.compare_versions(
            (self.__major, self.__minor, self.__patch),
            (other.major, other.minor, other.patch)
        )
        return result > 0

    def __init__(self, value, *args):
        """
        Build a ``Version`` instance providing the version of a software
        component as either:

        - A string representation of a semantic versioning 3-component number
          (at least 1);

        - From 1 to 3 integers representing, in that particular order, ``major``,
          ``minor``, and ``patch``;

        - A tuple of 3 integers ``(major, minor, patch)``.

        For example:

        ```python
        >>> Version("1")
        >>> Version("1.2")
        >>> Version("1.2.8")
        >>> Version(1)
        >>> Version(1, 2)
        >>> Version(1, 2, 8)
        >>> Version((1,))
        >>> Version((1, 2)
        >>> Version((1, 2, 8))
        ```


        :param value: Either a string representation of a software component
            of the following format: ``major.minor.patch``, or a tuple of 3
            integers `(major, minor, patch)`, or an integer representing the
            `major` component number.

        :param args: 1 or 2 integers for respectively the `minor`, and `patch`
            component numbers of the version, when the argument `value` passed
            to this function is an integer (the `major` component number).
        """
        if isinstance(value, string_types):
            self.__major, self.__minor, self.__patch, self.__prerelease, self.__build_metadata = self.__convert_string_to_version_component_numbers(value)
        elif isinstance(value, tuple):
            self.__major, self.__minor, self.__patch = value + (0,) * (3 - len(value))
            self.__prerelease = self.__build_metadata = None
        elif isinstance(value, int):
            minor_patch = args or ()
            minor_patch += (0,) * (2 - len(minor_patch))
            self.__major, self.__minor, self.__patch = (value,) + minor_patch
            self.__prerelease = self.__build_metadata = None
        else:
            raise ValueError(f"Invalid value '{value}'")

    def __le__(self, other: Version) -> bool:
        return self.__eq__(other) or self.__lt__(other)

    def __lt__(self, other: Version) -> bool:
        result = self.compare_versions(
            (self.__major, self.__minor, self.__patch),
            (other.major, other.minor, other.patch)
        )
        return result < 0

    def __repr__(self) -> str:
        return f'{self.__class__.__qualname__}({self.__major}, { self.__minor}, {self.__patch})'

    def __str__(self) -> str:
        return f'{self.__major}.{self.__minor}.{self.__patch}'

    @property
    def build_metadata(self) -> str or None:
        """
        Return the build metadata when defined.

        Build metadata is intended to track build maturity when preparing an
        application for a public release of any kind, including pre-releases.

        It can sometimes be helpful to include information such as:

        - The exact git hash of the commit used to produce the current build.

        - The release channel of the build.  For example, is this build a very
          experimental and potentially broken "nightly" release, a semi-stable
          "beta" release containing a few new experimental features, or a
          robust and well-tested "stable" release?

        - Whether the project as a whole is alpha-level quality (the entire
          project is pretty new and experimental).


        :return: The build metadata or ``None``.
        """
        return self.__build_metadata

    @staticmethod
    def compare_versions(
            this: (int, int, int),
            other: (int, int, int)
    ) -> int:
        """
        Compare a version to another.


        :param this: A tuple ``(major, minor, patch)``.

        :param other: One other tuple ``(major, minor, patch)``.


        :return: A negative integer if the first version passed to this
            function is below the second version passed, ``0`` if these two
            versions are equivalent, or a positive integer if the first
            version is above the second version.
        """
        return reduce(
            lambda a, b: 1 if a == 1 else -1 if a == -1 else b,
            # Use the function `starmap` instead ot the built-in function `map` as
            # Python 3 doesn't allow tuple parameter unpacking as `map(lambda (a, b): ...`.
            starmap(
                lambda a, b: 1 if a > b else -1 if a < b else 0,
                zip(this, other)
            )
        )

    @classmethod
    def __convert_string_to_version_component_numbers(
            cls,
            s: str,
            strict: bool = True
    ) -> (int, int, int, str, str):
        """
        Convert the string representation of a semantic versioning 3-component
        number into a tuple of integers `(major, minor, patch)`.

        If only 1-component number is given, the function returns `minor` and
        `patch` equal to `0`.


        :param s: A string representation of a semantic version 1- or
            3-component number.

        :param strict: Indicate whether the string ``s`` needs to strictly
            comply with Semantic Version.


        :return: A tuple of integers `(major. minor, patch, prerelease, build_metadata)`.


        :raise ValueError: If the string ``s`` doesn't comply with Semantic
            Versioning while the argument ``strict`` is ``true``.
        """
        if not isinstance(s, string_types):
            raise ValueError(f"The argument 's' is of the wrong type '{type(s)}'")

        if s is None and strict:
            raise ValueError(f"The argument 's' MUST NOT be null")

        match = re.match(cls.REGEX_PATTERN_SEMANTIC_VERSION, s.strip())
        if match:
            major = match.group('major')
            minor = match.group('minor')
            patch = match.group('patch')
            prerelease = match.group('prerelease')
            build_metadata = match.group('buildmetadata')

            return int(major), int(minor or 0), int(patch or 0), prerelease, build_metadata
        elif strict:
            raise ValueError(f"The argument 's' doesn't comply with Semantic Versioning")

    @classmethod
    def from_file(cls, path: str, file_name: str = None) -> Version:
        """
        Return a version written in a file.


        :param path: The absolute path of the version file.

        :param file_name: The name of the file where the version is written in.
            It defaults to {@link Version.DEFAULT_VERSION_FILE_NAME}.


        :return: An object {@link Version}.
        """
        version_file_path_name = os.path.join(path, file_name or cls.DEFAULT_VERSION_FILE_NAME)
        with io.open(version_file_path_name, mode='rt', encoding='utf-8') as fd:
            version = Version(fd.read())

        return version

    @staticmethod
    def from_string(s: str) -> Version or None:
        """
        Return the version corresponding to the specified string.


        :param s: A string representing a version.


        :return: The version corresponding to the string, or ``None`` if the
            argument ``s`` is null.
        """
        return s and Version(s)

    @property
    def major(self) -> int:
        return self.__major

    @property
    def minor(self) -> int:
        return self.__minor

    @property
    def patch(self) -> int:
        return self.__patch

    @property
    def prerelease(self) -> str or None:
        """
        Return the pre-release version when defined.

        A pre-release version indicates that the version is unstable and might
        not satisfy the intended compatibility requirements as denoted by its
        associated normal version.


        :return: The pre-release version or ``None``.
        """
        return self.__prerelease